# coding=utf-8

import asyncio

import pytest
import async_timeout

from xpaw.queue import FifoQueue, LifoQueue, PriorityQueue
from xpaw.http import HttpRequest


@pytest.mark.asyncio
async def test_fifo_queue():
    q = FifoQueue()
    with pytest.raises(asyncio.TimeoutError):
        with async_timeout.timeout(0.1):
            await q.pop()
    obj_list = [HttpRequest('1'), HttpRequest('2'), HttpRequest('3')]
    for o in obj_list:
        await q.push(o)
    for i in range(len(obj_list)):
        assert await q.pop() == obj_list[i]
    with pytest.raises(asyncio.TimeoutError):
        with async_timeout.timeout(0.1):
            await q.pop()


@pytest.mark.asyncio
async def test_lifo_queue():
    q = LifoQueue()
    with pytest.raises(asyncio.TimeoutError):
        with async_timeout.timeout(0.1):
            await q.pop()
    obj_list = [HttpRequest('1'), HttpRequest('2'), HttpRequest('3')]
    for o in obj_list:
        await q.push(o)
    for i in range(len(obj_list)):
        assert await q.pop() == obj_list[len(obj_list) - i - 1]
    with pytest.raises(asyncio.TimeoutError):
        with async_timeout.timeout(0.1):
            await q.pop()


@pytest.mark.asyncio
async def test_priority_queue():
    item1_1 = HttpRequest('1_1', priority=1)
    item1_2 = HttpRequest('1_2', priority=1)
    item2_1 = HttpRequest('2_1', priority=2)
    item2_2 = HttpRequest('2_2', priority=2)
    item3_1 = HttpRequest('3_1', priority=3)
    item3_2 = HttpRequest('3_2', priority=3)
    q = PriorityQueue()
    with pytest.raises(asyncio.TimeoutError):
        with async_timeout.timeout(0.1):
            await q.pop()
    await q.push(item2_1)
    await q.push(item1_1)
    await q.push(item3_1)
    await q.push(item1_2)
    await q.push(item2_2)
    await q.push(item3_2)
    assert await q.pop() is item3_1
    assert await q.pop() is item3_2
    assert await q.pop() is item2_1
    assert await q.pop() is item2_2
    assert await q.pop() is item1_1
    assert await q.pop() is item1_2
    with pytest.raises(asyncio.TimeoutError):
        with async_timeout.timeout(0.1):
            await q.pop()



# coding=utf-8

import asyncio
import logging
from asyncio import CancelledError
import time
import inspect

from .http import HttpRequest, HttpResponse
from .errors import IgnoreRequest, IgnoreItem, StopCrawler, ClientError, HttpError
from .downloader import DownloaderMiddlewareManager
from .spider import Spider, SpiderMiddlewareManager
from .eventbus import EventBus
from . import events
from .extension import ExtensionManager
from .item import BaseItem
from .pipeline import ItemPipelineManager
from .utils import load_object

log = logging.getLogger(__name__)


class Crawler:
    def __init__(self, config):
        self.config = config
        self.event_bus = EventBus()
        self.stats_collector = self._instance_from_crawler(self.config.get('stats_collector'))
        self.queue = self._instance_from_crawler(self.config.get('queue'))
        self.dupe_filter = self._instance_from_crawler(self.config.get('dupe_filter'))
        self.downloader = self._instance_from_crawler(self.config.get('downloader'))
        self.spider = self._instance_from_crawler(self.config.get('spider'))
        assert isinstance(self.spider, Spider), 'spider must inherit from the Spider class'
        log.info('Spider: %s', str(self.spider))
        self.downloader_middlware = DownloaderMiddlewareManager.from_crawler(self)
        log.info('Downloader middlewares: %s', self._log_objects(self.downloader_middlware.components))
        self.spider_middleware = SpiderMiddlewareManager.from_crawler(self)
        log.info('Spider middlewares: %s', self._log_objects(self.spider_middleware.components))
        self.item_pipeline = ItemPipelineManager.from_crawler(self)
        log.info('Item pipelines: %s', self._log_objects(self.item_pipeline.components))
        self.extension = ExtensionManager.from_crawler(self)
        log.info('Extensions: %s', self._log_objects(self.extension.components))

    async def start_requests(self):
        try:
            res = await self.spider_middleware.start_requests(self.spider)
        except CancelledError:
            raise
        except Exception:
            log.warning("Failed to get start requests", exc_info=True)
        else:
            return [r for r in res if isinstance(r, HttpRequest)]

    async def schedule(self, request):
        try:
            res = self.dupe_filter.is_duplicated(request)
            if inspect.iscoroutine(res):
                res = await res
            if not res:
                await self.event_bus.send(events.request_scheduled, request=request)
                await self.queue.push(request)
        except Exception:
            log.warning('Failed to schedule %s', request, exc_info=True)

    async def next_request(self):
        req = await self.queue.pop()
        return req

    async def fetch(self, req):
        try:
            resp = await self.downloader_middlware.fetch(req, self.downloader)
        except CancelledError:
            raise
        except Exception as e:
            if isinstance(e, IgnoreRequest):
                await self.event_bus.send(events.request_ignored, request=req, error=e)
            elif isinstance(e, (ClientError, HttpError)):
                log.debug('Failed to make %s: %s', req, e)
            else:
                log.warning("Failed to request %s", req, exc_info=True)
            await self.spider.request_error(req, e)
        else:
            await self._handle_response(resp)

    async def _handle_response(self, resp):
        if isinstance(resp, HttpRequest):
            await self.schedule(resp)
        elif isinstance(resp, HttpResponse):
            await self.event_bus.send(events.response_received, response=resp)
            try:
                result = await self.spider_middleware.parse(resp, self.spider)
            except CancelledError:
                raise
            except Exception as e:
                if isinstance(e, StopCrawler):
                    log.info('Request to stop crawler: %s', e)
                    raise
                elif isinstance(e, IgnoreRequest):
                    await self.event_bus.send(events.request_ignored, request=resp.request, error=e)
                else:
                    log.warning("Failed to parse %s", resp, exc_info=True)
            else:
                for r in result:
                    await self._handle_parsing_result(r)

    async def _handle_parsing_result(self, result):
        if isinstance(result, HttpRequest):
            await self.schedule(result)
        elif isinstance(result, (BaseItem, dict)):
            try:
                await self.item_pipeline.handle_item(result)
            except CancelledError:
                raise
            except Exception as e:
                if isinstance(e, IgnoreItem):
                    await self.event_bus.send(events.item_ignored, item=result, error=e)
                else:
                    log.warning("Failed to handle %s", result, exc_info=True)
            else:
                await self.event_bus.send(events.item_scraped, item=result)

    def _instance_from_crawler(self, cls_path):
        obj_cls = load_object(cls_path)
        if inspect.isclass(obj_cls):
            if hasattr(obj_cls, "from_crawler"):
                obj = obj_cls.from_crawler(self)
            else:
                obj = obj_cls()
        else:
            obj = obj_cls
        return obj

    @staticmethod
    def _log_objects(objects):
        if objects:
            return ''.join(['\n\t({}/{}) {}'.format(i + 1, len(objects), o) for i, o in enumerate(objects)])
        return ''


class CrawlerRunner:
    def __init__(self, crawler):
        self.crawler = crawler

        self._workers = None
        self._workers_done = None
        self._req_in_worker = None
        self._start_future = None
        self._supervisor_future = None
        self._is_running = False
        self._run_lock = None

    async def run(self):
        if self._is_running:
            return
        self._is_running = True
        await self._init()
        self._run_lock = asyncio.Future()
        await self._run_lock
        self._run_lock = None

    async def _init(self):
        await self.crawler.event_bus.send(events.crawler_start)
        self._supervisor_future = asyncio.ensure_future(self._supervisor())
        self._start_future = asyncio.ensure_future(self._generate_start_requests())
        downloader_clients = self.crawler.downloader.max_clients
        log.info("The maximum number of simultaneous clients: %s", downloader_clients)
        self._workers = []
        for i in range(downloader_clients):
            f = asyncio.ensure_future(self._download(i))
            self._workers.append(f)
        self._workers_done = set()
        self._req_in_worker = [None] * downloader_clients
        log.info('Crawler is initialized')

    def stop(self):
        if not self._is_running:
            return
        self._is_running = False
        asyncio.ensure_future(self._shutdown())

    async def _shutdown(self):
        log.info("Shutdown now")
        cancelled_futures = []
        if self._workers:
            for f in self._workers:
                f.cancel()
                cancelled_futures.append(f)
            self._workers = None
            self._workers_done = None
        if self._start_future:
            self._start_future.cancel()
            cancelled_futures.append(self._start_future)
            self._start_future = None
        if self._supervisor_future:
            self._supervisor_future.cancel()
            cancelled_futures.append(self._supervisor_future)
            self._supervisor_future = None
        # put back the unfinished requests
        if self._req_in_worker:
            for r in self._req_in_worker:
                if r:
                    await self.crawler.schedule(r)
            self._req_in_worker = None
        await self.crawler.event_bus.send(events.crawler_shutdown)
        # wait cancelled futures
        await asyncio.wait(cancelled_futures)
        log.info('Crawler is stopped')
        if self._run_lock:
            self._run_lock.set_result(True)

    async def _supervisor(self):
        while True:
            await asyncio.sleep(5)
            for i in range(len(self._workers)):
                f = self._workers[i]
                if f.done():
                    if i not in self._workers_done:
                        self._workers_done.add(i)
                        reason = "This future is cancelled" if f.cancelled() else str(f.exception())
                        log.error("Worker[%s] is shutdown: %s", i, reason)
                        self._req_in_worker[i] = None
            if self._all_done():
                break
        self.stop()

    def _all_done(self):
        if self._start_future.done() and len(self.crawler.queue) <= 0:
            no_active = True
            for i in range(len(self._workers)):
                if self._req_in_worker[i]:
                    no_active = False
                    break
            return no_active
        elif len(self._workers_done) == len(self._workers):
            log.error('No alive worker')
            return True
        return False

    async def _generate_start_requests(self):
        if hasattr(self.crawler.spider.start_requests, "cron_job"):
            tick = self.crawler.spider.start_requests.cron_tick
        else:
            tick = 0
        while True:
            t = time.time()
            reqs = await self.crawler.start_requests()
            for r in reqs:
                await self.crawler.schedule(r)
            if tick <= 0:
                break
            t = time.time() - t
            if t < tick:
                await asyncio.sleep(tick - t)

    async def _download(self, coro_id):
        while True:
            req = await self.crawler.next_request()
            log.debug("%s -> worker[%s]", req, coro_id)
            self._req_in_worker[coro_id] = req
            try:
                await self.crawler.fetch(req)
            except StopCrawler:
                self.stop()
                continue
            self._req_in_worker[coro_id] = None
            # check if it's all done
            if self._all_done():
                self.stop()



# coding=utf-8

import logging
import inspect
from asyncio import CancelledError
from urllib.parse import urlsplit
from asyncio import Semaphore

from tornado.httpclient import HTTPRequest, HTTPClientError
from tornado.curl_httpclient import CurlAsyncHTTPClient

import pycurl

from .middleware import MiddlewareManager
from .http import HttpRequest, HttpResponse, HttpHeaders
from .errors import ClientError, HttpError
from . import events
from .renderer import ChromeRenderer
from .utils import with_not_none_params

log = logging.getLogger(__name__)


class Downloader:
    def __init__(self, max_clients=100, renderer_cores=None):
        self._max_clients = max_clients
        self._http_client = CurlAsyncHTTPClient(max_clients=max_clients, force_instance=True)
        self._renderer = ChromeRenderer()
        if renderer_cores is None:
            renderer_cores = self._max_clients
        self._renderer_semaphore = Semaphore(renderer_cores)

    @classmethod
    def from_crawler(cls, crawler):
        config = crawler.config
        downloader = cls(**with_not_none_params(max_clients=config.getint('downloader_clients'),
                                                renderer_cores=config.getint('renderer_cores')))
        crawler.event_bus.subscribe(downloader.close, events.crawler_shutdown)
        return downloader

    @property
    def max_clients(self):
        return self._max_clients

    async def fetch(self, request):
        log.debug("HTTP request: %s", request)
        try:
            if request.render:
                async with self._renderer_semaphore:
                    response = await self._renderer.fetch(request)
            else:
                req = self._make_request(request)
                resp = await self._http_client.fetch(req)
                response = self._make_response(resp)
        except CancelledError:
            raise
        except HTTPClientError as e:
            if e.response is not None and e.response.code != 599:
                raise HttpError('{} {}'.format(e.response.code, e.message),
                                response=self._make_response(e.response))
            raise ClientError(e.message)
        except Exception as e:
            raise ClientError(e)
        log.debug("HTTP response: %s", response)
        return response

    def _make_request(self, request):
        kwargs = {'method': request.method,
                  'headers': request.headers,
                  'body': request.body,
                  'connect_timeout': request.timeout,  # FIXME
                  'request_timeout': request.timeout,  # FIXME
                  'follow_redirects': request.allow_redirects,
                  'validate_cert': request.verify_ssl}
        if request.auth is not None:
            auth_username, auth_password = request.auth
            kwargs['auth_username'] = auth_username
            kwargs['auth_password'] = auth_password
        if request.proxy is not None:
            s = urlsplit(request.proxy)
            if s.scheme:
                if s.scheme in ('http', 'socks4', 'socks5'):
                    proxy_host, proxy_port = s.hostname, s.port
                else:
                    raise ValueError('Unsupported proxy scheme: {}'.format(s.scheme))
                if s.scheme == 'socks5':
                    kwargs['prepare_curl_callback'] = prepare_curl_socks5
                elif s.scheme == 'socks4':
                    kwargs['prepare_curl_callback'] = prepare_curl_socks4
            else:
                proxy_host, proxy_port = request.proxy.split(':')
            kwargs['proxy_host'] = proxy_host
            kwargs['proxy_port'] = int(proxy_port)
        if request.proxy_auth is not None:
            proxy_username, proxy_password = request.proxy_auth
            kwargs['proxy_username'] = proxy_username
            kwargs['proxy_password'] = proxy_password
        return HTTPRequest(request.url, **kwargs)

    def _make_response(self, resp):
        return HttpResponse(resp.effective_url,
                            resp.code,
                            headers=resp.headers,
                            body=resp.body)

    def close(self):
        self._http_client.close()


def prepare_curl_socks5(curl):
    curl.setopt(pycurl.PROXYTYPE, pycurl.PROXYTYPE_SOCKS5)


def prepare_curl_socks4(curl):
    curl.setopt(pycurl.PROXYTYPE, pycurl.PROXYTYPE_SOCKS4)


class DownloaderMiddlewareManager(MiddlewareManager):
    def __init__(self, *middlewares):
        self._request_handlers = []
        self._response_handlers = []
        self._error_handlers = []
        super().__init__(*middlewares)

    def _add_middleware(self, middleware):
        super()._add_middleware(middleware)
        if hasattr(middleware, "handle_request"):
            self._request_handlers.append(middleware.handle_request)
        if hasattr(middleware, "handle_response"):
            self._response_handlers.insert(0, middleware.handle_response)
        if hasattr(middleware, "handle_error"):
            self._error_handlers.insert(0, middleware.handle_error)

    @classmethod
    def _middleware_list_from_config(cls, config):
        return cls._make_component_list('downloader_middlewares', config)

    async def fetch(self, request, downloader):
        try:
            res = await self._handle_request(request)
            if isinstance(res, HttpRequest):
                return res
            if res is None:
                res = await downloader.fetch(request)
        except CancelledError:
            raise
        except Exception as e:
            res = await self._handle_error(request, e)
            if isinstance(res, Exception):
                raise res
        if isinstance(res, HttpResponse):
            _res = await self._handle_response(request, res)
            if _res:
                res = _res
            # bind request
            res.request = request
        return res

    async def _handle_request(self, request):
        request.headers = self._make_request_headers(request.headers)
        for method in self._request_handlers:
            res = method(request)
            if inspect.iscoroutine(res):
                res = await res
            assert res is None or isinstance(res, (HttpRequest, HttpResponse)), \
                "Request handler must return None, HttpRequest or HttpResponse, got {}".format(type(res).__name__)
            if res:
                return res

    def _make_request_headers(self, headers):
        if isinstance(headers, HttpHeaders):
            return headers
        res = HttpHeaders()
        if isinstance(headers, dict):
            for k, v in headers.items():
                if isinstance(v, (tuple, list)):
                    for i in v:
                        res.add(k, i)
                else:
                    res.add(k, v)
        elif isinstance(headers, (tuple, list)):
            for k, v in headers:
                res.add(k, v)
        return res

    async def _handle_response(self, request, response):
        for method in self._response_handlers:
            res = method(request, response)
            if inspect.iscoroutine(res):
                res = await res
            assert res is None or isinstance(res, HttpRequest), \
                "Response handler must return None or HttpRequest, got {}".format(type(res).__name__)
            if res:
                return res

    async def _handle_error(self, request, error):
        for method in self._error_handlers:
            res = method(request, error)
            if inspect.iscoroutine(res):
                res = await res
            assert res is None or isinstance(res, (HttpRequest, HttpResponse)), \
                "Exception handler must return None, HttpRequest or HttpResponse, got {}".format(type(res).__name__)
            if res:
                return res
        return error



# coding=utf-8

import json

import pytest

from xpaw.http import HttpRequest, HttpResponse
from xpaw.downloader import Downloader, DownloaderMiddlewareManager
from xpaw import events
from xpaw.errors import HttpError
from xpaw.utils import make_url

from .crawler import Crawler


@pytest.mark.asyncio
async def test_basic_auth():
    downloader = Downloader()

    async def no_auth():
        req = HttpRequest("http://httpbin.org/basic-auth/user/passwd")
        with pytest.raises(HttpError) as e:
            await downloader.fetch(req)
        assert e.value.response.status == 401

    async def tuple_auth():
        req = HttpRequest("http://httpbin.org/basic-auth/user/passwd")
        req.auth = ('user', 'passwd')
        resp = await downloader.fetch(req)
        assert resp.status == 200

    await no_auth()
    await tuple_auth()


@pytest.mark.asyncio
async def test_params():
    downloader = Downloader()

    async def query_params():
        url = "http://httpbin.org/anything?key=value&none="
        resp = await downloader.fetch(HttpRequest(url))
        assert json.loads(resp.text)['args'] == {'key': 'value', 'none': ''}

    async def dict_params():
        resp = await downloader.fetch(
            HttpRequest(make_url("http://httpbin.org/get", params={'key': 'value', 'none': ''})))
        assert json.loads(resp.text)['args'] == {'key': 'value', 'none': ''}

    async def list_params():
        resp = await downloader.fetch(HttpRequest(make_url("http://httpbin.org/get",
                                                           params=[('list', '1'), ('list', '2')])))
        assert json.loads(resp.text)['args'] == {'list': ['1', '2']}

    await query_params()
    await dict_params()
    await list_params()


@pytest.mark.asyncio
async def test_headers():
    downloader = Downloader()
    headers = {'User-Agent': 'xpaw'}
    resp = await downloader.fetch(HttpRequest("http://httpbin.org/get",
                                              headers=headers))
    assert resp.status == 200
    data = json.loads(resp.text)['headers']
    assert 'User-Agent' in data and data['User-Agent'] == 'xpaw'


@pytest.mark.asyncio
async def test_body():
    downloader = Downloader()

    async def post_str():
        str_data = 'str data: 字符串数据'
        resp = await downloader.fetch(HttpRequest('http://httpbin.org/post',
                                                  'POST', body=str_data,
                                                  headers={'Content-Type': 'text/plain'}))
        assert resp.status == 200
        body = json.loads(resp.text)['data']
        assert body == str_data

    async def post_bytes():
        bytes_data = 'bytes data: 字节数据'
        resp = await downloader.fetch(HttpRequest('http://httpbin.org/post',
                                                  'POST', body=bytes_data.encode(),
                                                  headers={'Content-Type': 'text/plain'}))
        assert resp.status == 200
        body = json.loads(resp.text)['data']
        assert body == bytes_data

    await post_str()
    await post_bytes()


@pytest.mark.asyncio
async def test_allow_redirects():
    downloader = Downloader()

    resp = await downloader.fetch(HttpRequest(make_url('http://httpbin.org/redirect-to',
                                                       params={'url': 'http://python.org'})))
    assert resp.status // 100 == 2 and 'python.org' in resp.url

    with pytest.raises(HttpError) as e:
        await downloader.fetch(HttpRequest(make_url('http://httpbin.org/redirect-to',
                                                    params={'url': 'http://python.org'}),
                                           allow_redirects=False))
    assert e.value.response.status // 100 == 3


class FooDownloadermw:
    def __init__(self, d):
        self.d = d

    def open(self):
        self.d['open'] = ''

    def close(self):
        self.d['close'] = ''

    def handle_request(self, request):
        self.d['handle_request'] = request

    def handle_response(self, request, response):
        self.d['handle_response'] = (request, response)

    def handle_error(self, request, error):
        self.d['handle_error'] = (request, error)


class DummyDownloadermw:
    """
    no method
    """


class FooAsyncDownloaderMw(FooDownloadermw):
    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler.config['data'])

    async def handle_request(self, request):
        self.d['async_handle_request'] = request

    async def handle_response(self, request, response):
        self.d['async_handle_response'] = (request, response)

    async def handle_error(self, request, error):
        self.d['async_handle_error'] = (request, error)


@pytest.mark.asyncio
async def test_downloader_middleware_manager_handlers():
    data = {}
    crawler = Crawler(downloader_middlewares=[lambda d=data: FooDownloadermw(d),
                                              DummyDownloadermw,
                                              FooAsyncDownloaderMw],
                      default_downloader_middlewares=None,
                      data=data)
    downloadermw = DownloaderMiddlewareManager.from_crawler(crawler)
    request_obj = HttpRequest(None)
    response_obj = HttpResponse(None, None)
    error_obj = object()
    await crawler.event_bus.send(events.crawler_start)
    await downloadermw._handle_request(request_obj)
    await downloadermw._handle_response(request_obj, response_obj)
    await downloadermw._handle_error(request_obj, error_obj)
    await crawler.event_bus.send(events.crawler_shutdown)
    assert 'open' in data and 'close' in data
    assert data['handle_request'] is request_obj
    assert data['handle_response'][0] is request_obj and data['handle_response'][1] is response_obj
    assert data['handle_error'][0] is request_obj and data['handle_error'][1] is error_obj
    assert data['async_handle_request'] is request_obj
    assert data['async_handle_response'][0] is request_obj and data['async_handle_response'][1] is response_obj
    assert data['async_handle_error'][0] is request_obj and data['async_handle_error'][1] is error_obj

    data2 = {}
    crawler2 = Crawler(downloader_middlewares={lambda d=data2: FooDownloadermw(d): 0},
                       default_downloader_middlewares=None,
                       data=data2)
    downloadermw2 = DownloaderMiddlewareManager.from_crawler(crawler2)
    request_obj2 = HttpRequest(None)
    await crawler2.event_bus.send(events.crawler_start)
    await downloadermw2._handle_request(request_obj2)
    await crawler2.event_bus.send(events.crawler_shutdown)
    assert 'open' in data2 and 'close' in data2
    assert data2['handle_request'] is request_obj2

    crawler3 = Crawler(downloader_middlewares=None, default_downloader_middlewares=None, data={})
    DownloaderMiddlewareManager.from_crawler(crawler3)



# coding=utf-8

import random
import logging
import asyncio
from urllib.parse import urlsplit

from .errors import ClientError, NotEnabled, HttpError
from . import __version__
from .utils import with_not_none_params

log = logging.getLogger(__name__)


class RetryMiddleware:
    RETRY_ERRORS = (ClientError,)
    RETRY_HTTP_STATUS = (500, 502, 503, 504, 408, 429)

    def __init__(self, max_retry_times=3, retry_http_status=None):
        self._max_retry_times = max_retry_times
        self._retry_http_status = retry_http_status or self.RETRY_HTTP_STATUS

    def __repr__(self):
        cls_name = self.__class__.__name__
        return '{}(max_retry_times={}, retry_http_status={})' \
            .format(cls_name, repr(self._max_retry_times), repr(self._retry_http_status))

    @classmethod
    def from_crawler(cls, crawler):
        config = crawler.config
        if not config.getbool('retry_enabled'):
            raise NotEnabled
        return cls(**with_not_none_params(max_retry_times=config.getint('max_retry_times'),
                                          retry_http_status=config.getlist('retry_http_status')))

    def handle_response(self, request, response):
        for p in self._retry_http_status:
            if self.match_status(p, response.status):
                return self.retry(request, "HTTP status={}".format(response.status))

    @staticmethod
    def match_status(pattern, status):
        if isinstance(pattern, int):
            return pattern == status
        verse = False
        if pattern.startswith("!") or pattern.startswith("~"):
            verse = True
            pattern = pattern[1:]
        s = str(status)
        n = len(s)
        match = True
        if len(pattern) != n:
            match = False
        else:
            for i in range(n):
                if pattern[i] != "x" and pattern[i] != "X" and pattern[i] != s[i]:
                    match = False
                    break
        if verse:
            match = not match
        return match

    def handle_error(self, request, error):
        if isinstance(error, self.RETRY_ERRORS):
            return self.retry(request, error)
        if isinstance(error, HttpError):
            for p in self._retry_http_status:
                if self.match_status(p, error.response.status):
                    return self.retry(request, "HTTP status={}".format(error.response.status))

    def retry(self, request, reason):
        retry_times = request.meta.get('retry_times', 0) + 1
        if retry_times <= self._max_retry_times:
            log.debug('Retry %s (failed %s times): %s', request, retry_times, reason)
            retry_req = request.copy()
            retry_req.meta['retry_times'] = retry_times
            retry_req.dont_filter = True
            return retry_req
        else:
            log.debug('Give up retrying %s (failed %s times): %s', request, retry_times, reason)


class DefaultHeadersMiddleware:
    def __init__(self, default_headers=None):
        self._headers = default_headers or {}

    def __repr__(self):
        cls_name = self.__class__.__name__
        return '{}(default_headers={})'.format(cls_name, repr(self._headers))

    @classmethod
    def from_crawler(cls, crawler):
        default_headers = crawler.config.get("default_headers")
        if default_headers is None:
            raise NotEnabled
        return cls(default_headers=default_headers)

    def handle_request(self, request):
        for k, v in self._headers.items():
            request.headers.setdefault(k, v)


class ProxyMiddleware:
    def __init__(self, proxy):
        self._proxies = {'http': None, 'https': None}
        self._set_proxy(proxy)

    def __repr__(self):
        cls_name = self.__class__.__name__
        return '{}(proxy={})'.format(cls_name, repr(self._proxies))

    @classmethod
    def from_crawler(cls, crawler):
        proxy = crawler.config.get('proxy')
        if not proxy:
            raise NotEnabled
        return cls(proxy=proxy)

    def handle_request(self, request):
        if request.proxy is None:
            s = urlsplit(request.url)
            scheme = s.scheme or 'http'
            request.proxy = self._proxies.get(scheme)

    def _set_proxy(self, proxy):
        if isinstance(proxy, str):
            self._proxies['http'] = proxy
            self._proxies['https'] = proxy
        elif isinstance(proxy, dict):
            self._proxies.update(proxy)

    def _append_proxy(self, proxy, scheme):
        if isinstance(proxy, str):
            self._proxies[scheme].append(proxy)
        elif isinstance(proxy, (list, tuple)):
            for p in proxy:
                self._proxies[scheme].append(p)


class SpeedLimitMiddleware:
    def __init__(self, rate=1, burst=1):
        self._rate = rate
        self._burst = burst
        if self._rate <= 0:
            raise ValueError("rate must be greater than 0")
        if self._burst <= 0:
            raise ValueError('burst must be greater than 0')
        self._interval = 1.0 / self._rate
        self._bucket = self._burst
        self._semaphore = asyncio.Semaphore(self._burst)
        self._update_future = None

    def __repr__(self):
        cls_name = self.__class__.__name__
        return '{}(speed_limit_rate={}, speed_limit_burst={})'.format(cls_name, repr(self._rate), repr(self._burst))

    @classmethod
    def from_crawler(cls, crawler):
        config = crawler.config
        if config['speed_limit'] is None:
            raise NotEnabled
        return cls(**config['speed_limit'])

    async def handle_request(self, request):
        await self._semaphore.acquire()
        self._bucket -= 1

    async def _update_value(self):
        while True:
            await asyncio.sleep(self._interval)
            if self._bucket + 1 <= self._burst:
                self._bucket += 1
                self._semaphore.release()

    def open(self):
        self._update_future = asyncio.ensure_future(self._update_value())

    def close(self):
        if self._update_future:
            self._update_future.cancel()


class UserAgentMiddleware:
    DEVICE_TYPE = {'desktop', 'mobile'}
    BROWSER_TYPE = {'chrome'}
    BROWSER_DEFAULT_HEADERS = {
        'chrome': {
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8',
            'Accept-Encoding': 'gzip, deflate',
            'Accept-Language': 'zh,en;q=0.9',
            'Cache-Control': 'no-cache',
            'Connection': 'keep-alive',
            'Pragma': 'no-cache',
            'Upgrade-Insecure-Requests': '1'
        }
    }

    def __init__(self, user_agent=None, random_user_agent=False):
        self._device_type = None
        self._browser = None
        self._user_agent = self._make_user_agent(user_agent)
        self._is_random = random_user_agent

    def __repr__(self):
        cls_name = self.__class__.__name__
        return '{}(user_agent={}, random_user_agent={})'.format(cls_name, repr(self._user_agent), repr(self._is_random))

    @classmethod
    def from_crawler(cls, crawler):
        config = crawler.config
        user_agent = config.get('user_agent')
        random_user_agent = config.getbool('random_user_agent')
        return cls(user_agent=user_agent, random_user_agent=random_user_agent)

    def handle_request(self, request):
        if self._is_random:
            user_agent = self._gen_user_agent(self._device_type, self._browser)
        else:
            user_agent = self._user_agent
        request.headers['User-Agent'] = user_agent
        self._set_browser_default_headers(request)

    def _make_user_agent(self, ua):
        if ua and ua.startswith(':'):
            for i in ua[1:].split(','):
                if i in self.DEVICE_TYPE:
                    self._device_type = i
                elif i in self.BROWSER_TYPE:
                    self._browser = i
                else:
                    raise ValueError('Unknown user agent description: {}'.format(i))
            if self._browser is None:
                self._browser = 'chrome'
            if self._device_type is None:
                self._device_type = 'desktop'
            ua = self._gen_user_agent(self._device_type, self._browser)
        if not ua:
            ua = 'xpaw/{}'.format(__version__)
        return ua

    @staticmethod
    def _gen_user_agent(device, browser):
        if browser == 'chrome':
            chrome_version = '{}.0.{}.{}'.format(random.randint(51, 70),
                                                 random.randint(0, 9999), random.randint(0, 99))
            webkit = '{}.{}'.format(random.randint(531, 600), random.randint(0, 99))
            if device == 'desktop':
                os = 'Macintosh; Intel Mac OS X 10_13_6'
                return ('Mozilla/5.0 ({}) AppleWebKit/{} (KHTML, like Gecko) '
                        'Chrome/{} Safari/{}').format(os, webkit, chrome_version, webkit)
            elif device == 'mobile':
                os = 'iPhone; CPU iPhone OS 11_4_1 like Mac OS X'
                mobile = '15E{:03d}'.format(random.randint(0, 999))
                return ('Mozilla/5.0 ({}) AppleWebKit/{} (KHTML, like Gecko) '
                        'CriOS/{} Mobile/{} Safari/{}').format(os, webkit, chrome_version, mobile, webkit)

    def _set_browser_default_headers(self, request):
        if self._browser in self.BROWSER_DEFAULT_HEADERS:
            for k, v in self.BROWSER_DEFAULT_HEADERS[self._browser].items():
                request.headers.setdefault(k, v)



# coding=utf-8

import asyncio
from threading import Thread

from selenium.webdriver import Chrome, ChromeOptions

from .http import HttpResponse, HttpHeaders


class ChromeRenderer:
    async def fetch(self, request):
        lock = asyncio.Future()
        t = Thread(target=self._run_fetch_thread, args=(asyncio.get_event_loop(), lock, request))
        t.start()
        await lock
        result = lock.result()
        if isinstance(result, Exception):
            raise result
        return result

    def _run_fetch_thread(self, loop, lock, request):
        driver = Chrome(options=self.make_chrome_options(request))
        driver.set_page_load_timeout(request.timeout)
        driver.set_script_timeout(request.timeout)
        driver.implicitly_wait(request.timeout)
        try:
            driver.get(request.url)
            response = HttpResponse(driver.current_url, 200, body=driver.page_source.encode('utf-8'),
                                    headers=HttpHeaders(), request=request)
            loop.call_soon_threadsafe(lock.set_result, response)
        except Exception as e:
            loop.call_soon_threadsafe(lock.set_result, e)
        finally:
            driver.quit()

    @classmethod
    def make_chrome_options(cls, request):
        render_options = request.render if isinstance(request.render, dict) else {}
        chrome_options = ChromeOptions()
        chrome_options.add_argument('--headless')
        # anonymous
        chrome_options.add_argument('--incognito')
        chrome_options.add_argument('--ignore-certificate-errors')
        chrome_options.add_argument('--ignore-ssl-errors')
        chrome_options.add_argument('--disable-popup-blocking')
        chrome_options.add_argument('--disable-gpu')
        chrome_options.add_argument('--no-sandbox')
        prefs = {}
        if not render_options.get('show_images'):
            # disable images
            prefs.setdefault('profile.managed_default_content_settings.images', 2)
        if prefs:
            chrome_options.add_experimental_option('prefs', prefs)
        if 'mobile_emulation' in render_options:
            mobile_emulation = render_options['mobile_emulation']
            if not isinstance(mobile_emulation, dict):
                mobile_emulation = {'deviceName': 'iPhone 8'}
            chrome_options.add_experimental_option('mobileEmulation', mobile_emulation)
        return chrome_options



# coding=utf-8

import os
import re
import sys
import hashlib
import logging
from importlib import import_module
import string
from os.path import isfile
from urllib.parse import urlsplit, parse_qsl, urlencode, parse_qs
import cgi

from tornado.httputil import url_concat

PY35 = sys.version_info >= (3, 5)
PY36 = sys.version_info >= (3, 6)


def load_object(path):
    if isinstance(path, str):
        dot = path.rindex(".")
        module, name = path[:dot], path[dot + 1:]
        mod = import_module(module)
        return getattr(mod, name)
    return path


def configure_logger(name, config):
    log_level = config.get('log_level').upper()
    log_format = config.get('log_format')
    log_dateformat = config.get('log_dateformat')
    logger = logging.getLogger(name)
    logger.setLevel(log_level)
    log_file = config.get('log_file')
    if log_file:
        handler = logging.FileHandler(log_file)
    else:
        handler = logging.StreamHandler()
    handler.setLevel(log_level)
    formatter = logging.Formatter(log_format, log_dateformat)
    handler.setFormatter(formatter)
    logger.handlers.clear()
    logger.addHandler(handler)
    return logger


def configure_tornado_logger(handlers):
    log = logging.getLogger('tornado')
    if log.handlers:
        return
    log.handlers = handlers
    log.setLevel('WARNING')


def request_fingerprint(request):
    sha1 = hashlib.sha1()
    sha1.update(to_bytes(request.method))
    res = urlsplit(request.url)
    queries = parse_qsl(res.query)
    queries.sort()
    final_query = urlencode(queries)
    sha1.update(to_bytes('{}://{}{}:{}?{}'.format(res.scheme,
                                                  '' if res.hostname is None else res.hostname,
                                                  res.path,
                                                  80 if res.port is None else res.port,
                                                  final_query)))
    sha1.update(request.body or b'')
    return sha1.hexdigest()


def to_bytes(data, encoding=None):
    if isinstance(data, bytes):
        return data
    if isinstance(data, str):
        return data.encode(encoding or "utf-8")
    raise TypeError("Need bytes or str, got {}".format(type(data).__name__))


def render_template_file(path, **kwargs):
    if path.endswith(".tmpl"):
        with open(path, "rb") as f:
            raw = f.read().decode("utf-8")
        content = string.Template(raw).substitute(**kwargs)
        render_path = path[:-len(".tmpl")]
        with open(render_path, "wb") as f:
            f.write(content.encode("utf-8"))
        os.remove(path)


_camelcase_invalid_chars = re.compile(r'[^a-zA-Z\d]')


def string_camelcase(s):
    return _camelcase_invalid_chars.sub('', s.title())


async def iterable_to_list(gen):
    res = []
    if gen is not None:
        if hasattr(gen, '__aiter__'):
            async for r in gen:
                res.append(r)
        else:
            for r in gen:
                res.append(r)
    return res


def cmp(a, b):
    return (a > b) - (a < b)


def daemonize():
    if os.fork():
        os._exit(0)
    os.setsid()
    if os.fork():
        os._exit(0)
    os.umask(0o22)
    os.closerange(0, 3)
    fd_null = os.open(os.devnull, os.O_RDWR)
    if fd_null != 0:
        os.dup2(fd_null, 0)
    os.dup2(fd_null, 1)
    os.dup2(fd_null, 2)


def load_config(fname):
    if fname is None or not isfile(fname):
        raise ValueError('{} is not a file'.format(fname))
    code = compile(open(fname, 'rb').read(), fname, 'exec')
    cfg = {
        "__builtins__": __builtins__,
        "__name__": "__config__",
        "__file__": fname,
        "__doc__": None,
        "__package__": None
    }
    exec(code, cfg, cfg)
    return cfg


def iter_settings(config):
    for key, value in config.items():
        if not key.startswith('_'):
            yield key, value


def get_encoding_from_content_type(content_type):
    if content_type:
        content_type, params = cgi.parse_header(content_type)
        if "charset" in params:
            return params["charset"]


_charset_flag = re.compile(r"""<meta.*?charset=["']*(.+?)["'>]""", flags=re.I)
_pragma_flag = re.compile(r"""<meta.*?content=["']*;?charset=(.+?)["'>]""", flags=re.I)
_xml_flag = re.compile(r"""^<\?xml.*?encoding=["']*(.+?)["'>]""")


def get_encoding_from_content(content):
    if isinstance(content, bytes):
        content = content.decode("ascii", errors="ignore")
    elif not isinstance(content, str):
        raise ValueError("content should be bytes or str")
    s = _charset_flag.search(content)
    if s:
        return s.group(1).strip()
    s = _pragma_flag.search(content)
    if s:
        return s.group(1).strip()
    s = _xml_flag.search(content)
    if s:
        return s.group(1).strip()


def make_url(url, params=None):
    args = []
    if isinstance(params, dict):
        for k, v in params.items():
            if isinstance(v, (tuple, list)):
                for i in v:
                    args.append((k, i))
            else:
                args.append((k, v))
    elif isinstance(params, (tuple, list)):
        for k, v in params:
            args.append((k, v))
    return url_concat(url, args)


def get_params_in_url(url):
    return parse_qs(urlsplit(url).query)


def with_not_none_params(**kwargs):
    params = {}
    for k, v in kwargs.items():
        if v is not None:
            params[k] = v
    return params



# coding=utf-8

import inspect

from tornado.httputil import HTTPHeaders

from .utils import get_encoding_from_content, get_encoding_from_content_type, make_url

HttpHeaders = HTTPHeaders


class HttpRequest:
    def __init__(self, url, method="GET", body=None, params=None, headers=None, proxy=None,
                 timeout=20, verify_ssl=False, allow_redirects=True, auth=None, proxy_auth=None,
                 priority=None, dont_filter=False, callback=None, errback=None, meta=None,
                 render=None):
        """
        Construct an HTTP request.
        """
        self.url = make_url(url, params)
        self.method = method
        self.body = body
        self.headers = headers
        self.proxy = proxy
        self.timeout = timeout
        self.verify_ssl = verify_ssl
        self.allow_redirects = allow_redirects
        self.auth = auth
        self.proxy_auth = proxy_auth
        self.priority = priority
        self.dont_filter = dont_filter
        self.callback = callback
        self.errback = errback
        self._meta = dict(meta) if meta else {}
        self.render = render

    def __str__(self):
        return '<{}, {}>'.format(self.method, self.url)

    __repr__ = __str__

    @property
    def meta(self):
        return self._meta

    def copy(self):
        return self.replace()

    def replace(self, **kwargs):
        for i in ["url", "method", "body", "headers", "proxy",
                  "timeout", "verify_ssl", "allow_redirects", "auth", "proxy_auth",
                  "priority", "dont_filter", "callback", "errback", "meta",
                  "render"]:
            kwargs.setdefault(i, getattr(self, i))
        return type(self)(**kwargs)

    def to_dict(self):
        callback = self.callback
        if inspect.ismethod(callback):
            callback = callback.__name__
        errback = self.errback
        if inspect.ismethod(errback):
            errback = errback.__name__
        d = {
            'url': self.url,
            'method': self.method,
            'body': self.body,
            'headers': self.headers,
            'proxy': self.proxy,
            'timeout': self.timeout,
            'verify_ssl': self.verify_ssl,
            'allow_redirects': self.allow_redirects,
            'auth': self.auth,
            'proxy_auth': self.proxy_auth,
            'priority': self.priority,
            'dont_filter': self.dont_filter,
            'callback': callback,
            'errback': errback,
            'meta': self.meta,
            'render': self.render
        }
        return d

    @classmethod
    def from_dict(cls, d):
        return cls(**d)


class HttpResponse:
    def __init__(self, url, status, body=None, headers=None,
                 request=None, encoding=None):
        """
        Construct an HTTP response.
        """
        self.url = url
        self.status = status
        self.body = body
        self.headers = headers
        self.request = request
        self._encoding = encoding

    def __str__(self):
        return '<{}, {}>'.format(self.status, self.url)

    __repr__ = __str__

    @property
    def encoding(self):
        if self._encoding:
            return self._encoding
        encoding = get_encoding_from_content_type(self.headers.get("Content-Type"))
        if not encoding and self.body:
            encoding = get_encoding_from_content(self.body)
        return encoding or 'utf-8'

    @encoding.setter
    def encoding(self, value):
        self._encoding = value

    @property
    def text(self):
        if hasattr(self, "_text") and self._text:
            return self._text
        if not self.body:
            return ""
        if isinstance(self.body, bytes):
            self._text = self.body.decode(self.encoding, errors="replace")
        else:
            self._text = self.body
        return self._text

    @property
    def meta(self):
        if self.request:
            return self.request.meta

    def copy(self):
        return self.replace()

    def replace(self, **kwargs):
        for i in ["url", "status", "body", "headers", "request"]:
            kwargs.setdefault(i, getattr(self, i))
        return type(self)(**kwargs)



# coding=utf-8

from .http import HttpRequest, HttpResponse, HttpHeaders
from .downloader import Downloader
from .spider import Spider
from .selector import Selector
from .item import Item, Field
from .run import run_spider, run_spider_project, make_requests
from .decorator import every

__all__ = ['HttpRequest', 'HttpResponse', 'HttpHeaders',
           'Downloader',
           'Spider',
           'Selector',
           'Item', 'Field',
           'run_spider', 'run_spider_project', 'make_requests',
           'every']

__version__ = '0.11.1'



# coding=utf-8

import pytest

from xpaw.pipeline import ItemPipelineManager
from xpaw import events
from .crawler import Crawler


class FooItemPipeline:
    def __init__(self, d):
        self.d = d

    def handle_item(self, item):
        self.d['handle_item'] = item

    def open(self):
        self.d['open'] = ''

    def close(self):
        self.d['close'] = ''


class DummyItemPipeline:
    """
    no method
    """


class FooAsyncItemPipeline(FooItemPipeline):
    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler.config['data'])

    @pytest.mark.asyncio
    async def handle_item(self, item):
        self.d['async_handle_item'] = item


@pytest.mark.asyncio
async def test_item_pipeline_manager():
    data = {}
    crawler = Crawler(item_pipelines=[lambda d=data: FooItemPipeline(d),
                                      DummyItemPipeline,
                                      FooAsyncItemPipeline],
                      default_item_pipelines=None,
                      data=data)
    pipeline = ItemPipelineManager.from_crawler(crawler)
    obj = object()
    await crawler.event_bus.send(events.crawler_start)
    await pipeline.handle_item(obj)
    await crawler.event_bus.send(events.crawler_shutdown)
    assert 'open' in data and 'close' in data
    assert data['handle_item'] is obj and data['async_handle_item'] is obj

    crawler3 = Crawler(item_pipelines=None, default_item_pipelines=None, data={})
    ItemPipelineManager.from_crawler(crawler3)



# coding=utf-8

from xpaw import Spider, HttpRequest, Selector, run_spider


class RenderingSpider(Spider):
    def start_requests(self):
        yield HttpRequest('http://quotes.toscrape.com/js/', callback=self.parse, render=True)

    def parse(self, response):
        selector = Selector(response.text)
        for quote in selector.css('div.quote'):
            text = quote.css('span.text')[0].text
            author = quote.css('small.author')[0].text
            self.log(author + ": " + text)


if __name__ == '__main__':
    run_spider(RenderingSpider)



# coding=utf-8

import pytest

from xpaw.extension import ExtensionManager
from xpaw import events
from .crawler import Crawler


class FooExtension:
    def __init__(self, d):
        self.d = d

    def open(self):
        self.d['open'] = ''

    def close(self):
        self.d['close'] = ''


class DummyExtension:
    """
    no method
    """


@pytest.mark.asyncio
async def test_extension_manager():
    data = {}
    crawler = Crawler(extensions=[lambda d=data: FooExtension(d),
                                  DummyExtension],
                      default_extensions=None,
                      data=data)
    extensions = ExtensionManager.from_crawler(crawler)
    await crawler.event_bus.send(events.crawler_start)
    await crawler.event_bus.send(events.crawler_shutdown)
    assert 'open' in data and 'close' in data

    crawler3 = Crawler(extensions=None, default_extensions=None, data={})
    ExtensionManager.from_crawler(crawler3)



# coding=utf-8

import pytest

from xpaw.decorator import every


def test_every_value_error():
    with pytest.raises(ValueError):
        @every()
        def func():
            pass


def test_every():
    @every(hours=1)
    def func_hours():
        pass

    assert func_hours.cron_job is True and func_hours.cron_tick == 3600

    @every(minutes=1)
    def func_minutes():
        pass

    assert func_minutes.cron_job is True and func_minutes.cron_tick == 60

    @every(seconds=1)
    def func_seconds():
        pass

    assert func_seconds.cron_job is True and func_seconds.cron_tick == 1

    @every(hours=1, minutes=1, seconds=1)
    def func():
        pass

    assert func.cron_job is True and func.cron_tick == 3661



# coding=utf-8

from xpaw import Spider, HttpRequest, run_spider
from xpaw.errors import HttpError, ClientError


class ErrorHandlingSpider(Spider):
    start_urls = [
        "http://www.python.org/",  # 200 OK
        "http://www.httpbin.org/status/404",  # 404 Not Found
        "http://www.httpbin.org/status/500",  # 500 Service Not Available
        "http://unknown/",  # ClientError
    ]

    def start_requests(self):
        for url in self.start_urls:
            yield HttpRequest(url, errback=self.handle_error)

    def parse(self, response):
        self.logger.info('Successful response: %s', response)

    def handle_error(self, request, error):
        if isinstance(error, HttpError):
            response = error.response
            self.logger.error('HttpError on %s: HTTP status=%s', request.url, response.status)
        elif isinstance(error, ClientError):
            self.logger.error('ClientError on %s: %s', request.url, error)


if __name__ == '__main__':
    run_spider(ErrorHandlingSpider, retry_enabled=False)



# coding=utf-8

from os.path import join

import pytest

from xpaw.spider import Spider
from xpaw.cmdline import main
from xpaw.run import run_spider, run_spider_project, make_requests
from xpaw.http import HttpRequest, HttpResponse
from xpaw.errors import ClientError, HttpError


def test_run_spider_project(tmpdir):
    proj_name = 'test_run_spider_project'
    proj_dir = join(str(tmpdir), proj_name)
    main(argv=['xpaw', 'init', proj_dir])
    run_spider_project(proj_dir, log_level='DEBUG')


def test_run_spider_project_with_bad_config(tmpdir, capsys):
    proj_dir = join(str(tmpdir))
    config_file = join(proj_dir, 'config.py')
    with open(config_file, 'w') as f:
        f.write('bad config')
    with pytest.raises(SyntaxError):
        run_spider_project(proj_dir, log_level='DEBUG')
    _, _ = capsys.readouterr()


def test_failed_to_create_crawler(tmpdir, capsys):
    proj_dir = join(str(tmpdir))
    with pytest.raises(Exception):
        run_spider_project(proj_dir, log_level='DEBUG')
    _, _ = capsys.readouterr()


class DummySpider(Spider):
    def start_requests(self):
        pass

    def parse(self, response):
        pass


def test_run_spider():
    run_spider(DummySpider, log_level='DEBUG')


def test_make_requests():
    requests = [None, 'http://unknonw',
                'http://python.org/', HttpRequest('http://python.org'),
                'http://httpbin.org/status/404']
    results = make_requests(requests, log_level='DEBUG')
    assert len(results) == len(requests)
    assert results[0] is None
    assert isinstance(results[1], ClientError)
    assert isinstance(results[2], HttpResponse) and results[2].status == 200
    assert isinstance(results[3], HttpResponse) and results[3].status == 200
    assert isinstance(results[4], HttpError) and results[4].response.status == 404



# coding=utf-8


def every(hours=None, minutes=None, seconds=None):
    def wrapper(func):
        func.cron_job = True
        func.cron_tick = hours * 3600 + minutes * 60 + seconds
        return func

    if hours is None and minutes is None and seconds is None:
        raise ValueError('At least one of the parameters (hours, minutes and seconds) is not none')
    if hours is None:
        hours = 0
    if minutes is None:
        minutes = 0
    if seconds is None:
        seconds = 0
    return wrapper



# coding=utf-8

import pytest

from xpaw.spider import SpiderMiddlewareManager, Spider
from xpaw import events
from .crawler import Crawler


class FooSpidermw:
    def __init__(self, d):
        self.d = d

    def open(self):
        self.d['open'] = ''

    def close(self):
        self.d['close'] = ''

    def handle_start_requests(self, result):
        res = []
        for r in result:
            self.d['handle_start_requests'] = r
            res.append(r)
        return res

    def handle_input(self, response):
        self.d['handle_input'] = response

    def handle_output(self, response, result):
        res = []
        for r in result:
            self.d['handle_output'] = (response, r)
            res.append(r)
        return res

    def handle_error(self, response, error):
        self.d['handle_error'] = (response, error)


class DummySpidermw:
    """
    no method
    """


class FooAsyncSpiderMw(FooSpidermw):
    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler.config['data'])

    @pytest.mark.asyncio
    async def handle_start_requests(self, result):
        res = []
        for r in result:
            self.d['async_handle_start_requests'] = r
            res.append(r)
        return res

    @pytest.mark.asyncio
    async def handle_input(self, response):
        self.d['async_handle_input'] = response

    @pytest.mark.asyncio
    async def handle_output(self, response, result):
        res = []
        for r in result:
            self.d['async_handle_output'] = (response, r)
            res.append(r)
        return res

    @pytest.mark.asyncio
    async def handle_error(self, response, error):
        self.d['async_handle_error'] = (response, error)


@pytest.mark.asyncio
async def test_spider_middleware_manager_handlers():
    data = {}
    crawler = Crawler(spider_middlewares=[lambda d=data: FooSpidermw(d),
                                          DummySpidermw,
                                          FooAsyncSpiderMw],
                      default_spider_middlewares=None,
                      data=data)
    spidermw = SpiderMiddlewareManager.from_crawler(crawler)
    response_obj = object()
    result_obj = object()
    error_obj = object()
    await crawler.event_bus.send(events.crawler_start)
    await spidermw._handle_start_requests((result_obj,))
    await spidermw._handle_input(response_obj)
    await spidermw._handle_output(response_obj, (result_obj,))
    await spidermw._handle_error(response_obj, error_obj)
    await crawler.event_bus.send(events.crawler_shutdown)
    assert 'open' in data and 'close' in data
    assert data['handle_start_requests'] is result_obj
    assert data['handle_input'] is response_obj
    assert data['handle_output'][0] is response_obj and data['handle_output'][1] is result_obj
    assert data['handle_error'][0] is response_obj and data['handle_error'][1] is error_obj
    assert data['async_handle_start_requests'] is result_obj
    assert data['async_handle_input'] is response_obj
    assert data['async_handle_output'][0] is response_obj and data['async_handle_output'][1] is result_obj
    assert data['async_handle_error'][0] is response_obj and data['async_handle_error'][1] is error_obj

    data2 = {}
    crawler2 = Crawler(spider_middlewares={lambda d=data2: FooSpidermw(d): 0},
                       default_spider_middlewares=None,
                       data=data2)
    spidermw2 = SpiderMiddlewareManager.from_crawler(crawler2)
    response_obj2 = object()
    await crawler2.event_bus.send(events.crawler_start)
    await spidermw2._handle_input(response_obj2)
    await crawler2.event_bus.send(events.crawler_shutdown)
    assert 'open' in data2 and 'close' in data2
    assert data2['handle_input'] is response_obj2

    crawler3 = Crawler(spider_middlewares=None, default_spider_middlewares=None, data={})
    SpiderMiddlewareManager.from_crawler(crawler3)


class FooSpider(Spider):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data = self.config['data']

    def open(self):
        super().open()
        self.data['open'] = ''

    def close(self):
        super().close()
        self.data['close'] = ''


@pytest.mark.asyncio
async def test_spider():
    data = {}
    crawler = Crawler(data=data)
    spider = FooSpider.from_crawler(crawler)
    await crawler.event_bus.send(events.crawler_start)
    with pytest.raises(NotImplementedError):
        spider.start_requests()
    with pytest.raises(NotImplementedError):
        spider.parse(None)
    await crawler.event_bus.send(events.crawler_shutdown)
    assert 'open' in data and 'close' in data



# coding=utf-8

import pytest

from xpaw.http import HttpRequest, HttpResponse
from xpaw.downloader_middlewares import *
from xpaw.errors import ClientError, NotEnabled
from .crawler import Crawler


class TestDefaultHeadersMiddleware:
    def test_handle_request(self):
        default_headers = {"User-Agent": "xpaw", "Connection": "keep-alive"}
        mw = DefaultHeadersMiddleware.from_crawler(Crawler(default_headers=default_headers))
        req = HttpRequest("http://example.com", headers={"Connection": "close"})
        mw.handle_request(req)
        assert req.headers == {"User-Agent": "xpaw", "Connection": "close"}

    def test_not_enabled(self):
        with pytest.raises(NotEnabled):
            DefaultHeadersMiddleware.from_crawler(Crawler(default_headers=None))


class Random:
    def __init__(self):
        self.iter = 0

    def choice(self, seq):
        res = seq[self.iter % len(seq)]
        self.iter += 1
        return res


class TestProxyMiddleware:
    def test_proxy_str(self):
        proxy = '127.0.0.1:3128'
        mw = ProxyMiddleware.from_crawler(Crawler(proxy=proxy))
        req = HttpRequest("http://example.com")
        mw.handle_request(req)
        assert req.proxy == proxy

    def test_proxy_dict(self):
        proxy_dict = {'http': '127.0.0.1:3128', 'https': '127.0.0.1:3129'}
        mw = ProxyMiddleware.from_crawler(Crawler(proxy=proxy_dict))
        req_list = []
        for i in ['http://example.com', 'https://example.com']:
            req_list.append(HttpRequest(i))
        res = ['127.0.0.1:3128', '127.0.0.1:3129']
        for i in range(len(req_list)):
            mw.handle_request(req_list[i])
            assert req_list[i].proxy == res[i]

    @pytest.mark.asyncio
    async def test_not_enabled(self):
        with pytest.raises(NotEnabled):
            ProxyMiddleware.from_crawler(Crawler())


class TestRetryMiddleware:
    def test_handle_reponse(self):
        mw = RetryMiddleware.from_crawler(Crawler(retry_http_status=(500,), max_retry_times=3))
        req = HttpRequest("http://example.com")
        resp = HttpResponse("http://example.com", 502)
        assert mw.handle_response(req, resp) is None
        req2 = HttpRequest("http://example.com")
        resp2 = HttpResponse("http://example.com", 500)
        retry_req2 = mw.handle_response(req2, resp2)
        assert retry_req2.meta['retry_times'] == 1
        assert str(retry_req2.url) == str(req2.url)
        req3 = HttpRequest("http://example.com")
        resp3 = HttpResponse("http://example.com", 500)
        req3.meta['retry_times'] = 2
        retry_req3 = mw.handle_response(req3, resp3)
        assert retry_req3.meta['retry_times'] == 3
        assert str(retry_req3.url) == str(req3.url)
        req4 = HttpRequest("http://example.com")
        req4.meta['retry_times'] = 3
        resp4 = HttpResponse("http://example.com", 500)
        assert mw.handle_response(req4, resp4) is None

    def test_handle_error(self):
        mw = RetryMiddleware.from_crawler(Crawler())
        req = HttpRequest("http://example.com")
        err = ValueError()
        assert mw.handle_error(req, err) is None
        for err in [ClientError()]:
            retry_req = mw.handle_error(req, err)
            assert isinstance(retry_req, HttpRequest) and str(retry_req.url) == str(req.url)

    def test_retry(self):
        max_retry_times = 2
        mw = RetryMiddleware.from_crawler(Crawler(max_retry_times=max_retry_times,
                                                  retry_http_status=(500,)))
        req = HttpRequest("http://example.com")
        for i in range(max_retry_times):
            retry_req = mw.retry(req, "")
            assert isinstance(retry_req, HttpRequest) and str(retry_req.url) == str(req.url)
            req = retry_req
        assert mw.retry(req, "") is None

    def test_not_enabled(self):
        with pytest.raises(NotEnabled):
            RetryMiddleware.from_crawler(Crawler(retry_enabled=False))

    def test_match_status(self):
        assert RetryMiddleware.match_status("200", 200) is True
        assert RetryMiddleware.match_status(200, 200) is True
        assert RetryMiddleware.match_status("2xX", 201) is True
        assert RetryMiddleware.match_status("40x", 403) is True
        assert RetryMiddleware.match_status("40X", 403) is True
        assert RetryMiddleware.match_status("50x", 403) is False
        assert RetryMiddleware.match_status("~20X", 200) is False
        assert RetryMiddleware.match_status("!20x", 400) is True
        assert RetryMiddleware.match_status("0200", 200) is False


class TestSpeedLimitMiddleware:
    @pytest.mark.asyncio
    async def test_value_error(self):
        with pytest.raises(ValueError):
            SpeedLimitMiddleware.from_crawler(Crawler(speed_limit={'rate': 0, 'burst': 1}))
        with pytest.raises(ValueError):
            SpeedLimitMiddleware.from_crawler(Crawler(speed_limit={'rate': 1, 'burst': 0}))

    @pytest.mark.asyncio
    async def test_not_enabled(self):
        with pytest.raises(NotEnabled):
            SpeedLimitMiddleware.from_crawler(Crawler())

    @pytest.mark.asyncio
    async def test_handle_request(self):
        class Counter:
            def __init__(self):
                self.n = 0

            def inc(self):
                self.n += 1

        async def processor():
            while True:
                await mw.handle_request(None)
                counter.inc()

        counter = Counter()
        mw = SpeedLimitMiddleware.from_crawler(Crawler(speed_limit={'rate': 1000, 'burst': 5}))
        futures = []
        for i in range(100):
            futures.append(asyncio.ensure_future(processor()))
        mw.open()
        await asyncio.sleep(0.1)
        mw.close()
        for f in futures:
            assert f.cancel() is True
        await asyncio.sleep(0.01)
        assert counter.n <= 105


class TestUserAgentMiddleware:
    def test_static_user_agent(self):
        user_agent = 'test user agent'
        mw = UserAgentMiddleware.from_crawler(Crawler(user_agent=user_agent))
        req = HttpRequest('http://example.com', headers={})
        mw.handle_request(req)
        assert req.headers.get('User-Agent') == user_agent

    def test_gen_user_agent(self):
        mw = UserAgentMiddleware.from_crawler(Crawler(user_agent=':desktop,chrome'))
        req = HttpRequest('http://example.com', headers={})
        mw.handle_request(req)
        assert 'Chrome' in req.headers.get('User-Agent')

        mw2 = UserAgentMiddleware.from_crawler(Crawler(user_agent=':mobile,chrome'))
        req2 = HttpRequest('http://example.com', headers={})
        mw2.handle_request(req2)
        assert 'CriOS' in req2.headers.get('User-Agent') and 'Mobile' in req2.headers.get('User-Agent')

    def test_unknown_user_agent_description(self):
        with pytest.raises(ValueError):
            UserAgentMiddleware.from_crawler(Crawler(user_agent=':unknown'))

    def test_random_user_agent(self):
        mw = UserAgentMiddleware.from_crawler(Crawler(random_user_agent=True))
        req = HttpRequest('http://example.com', headers={})
        req2 = HttpRequest('http://example.com', headers={})
        mw.handle_request(req)
        mw.handle_request(req2)
        assert 'User-Agent' in req.headers
        assert req.headers.get('User-Agent') != req2.headers.get('User-Agent')
        assert 'Chrome' in req.headers.get('User-Agent')

    def test_random_user_agent2(self):
        mw = UserAgentMiddleware.from_crawler(Crawler(user_agent=':mobile', random_user_agent=True))
        for i in range(30):
            req = HttpRequest('http://example.com', headers={})
            mw.handle_request(req)
            assert 'User-Agent' in req.headers
            assert 'CriOS' in req.headers.get('User-Agent') and 'Mobile' in req.headers.get('User-Agent')



# coding=utf-8

from xpaw.eventbus import EventBus
from xpaw.config import Config, DEFAULT_CONFIG


class Crawler:
    def __init__(self, **kwargs):
        self.event_bus = EventBus()
        self.config = Config(DEFAULT_CONFIG, **kwargs)
